import numpy as np
import pandas as pd
from itertools import count
import collections
from fbprophet import Prophet
from stldecompose import decompose, forecast
from stldecompose.forecast_funcs import (naive,
                                         drift, 
                                         mean, 
                                         seasonal_naive)

from tsforest.config import prophet_kwargs,prophet_kwargs_extra

import warnings
warnings.filterwarnings("ignore", message="invalid value encountered in subtract")
warnings.simplefilter(action='ignore', category=FutureWarning)
import logging
logging.getLogger("fbprophet").setLevel(logging.WARNING)


def moving_average(data, window_size):
    """ 
    Computes moving average using discrete linear convolution 
    of two one dimensional sequences.
    Parameters
    ----------
            data (pandas.Series): independent variable
            window_size (int): rolling window size
    Returns
    ---------
            np.ndarray of linear convolution
    """
    window = np.ones(int(window_size))/float(window_size)
    return np.convolve(data, window, 'same')

def anomaly_detector(y, window_size=10, sigma=2.0):
    """ 
    Identifies anomalies in TS data through rolling std

    Parameters
    ----------
    y: pandas.Series 
        independent variable
    window_size: int 
        rolling window size
    sigma : int 
        value for standard deviation
    Returns
    --------
    a dict (dict of 'standard_deviation': int, 'anomalies_dict': (index: value))
    containing information about the points indentified as anomalies
    """
    avg = moving_average(y, window_size)
    avg_list = avg.tolist()
    residual = y - avg
    # calculate the variation in the distribution of the residual
    testing_std = residual.rolling(window_size).std()
    testing_std_as_df = pd.DataFrame(testing_std)
    rolling_std = testing_std_as_df.replace(np.nan,
                                  testing_std_as_df.loc[window_size - 1]).round(3).iloc[:,0].tolist()
    ret = collections.OrderedDict([(index, y_i) for 
                                   index, y_i, avg_i, rs_i in zip(count(),y,avg_list,rolling_std) 
                                   if (y_i > avg_i + (sigma * rs_i)) | (y_i < avg_i - (sigma * rs_i))])
    return ret
    
def ts_cleaning(df):
    """
    Parameters
    -----------
    df: pd.DataFrame
        Dataframe with columns "ds" and "y" containing the TS
    Returns
    -----------
    pd.Dataframe
        Dataframe in the same format as input, with outliers removed 
        from the TS
    """
    df = df.copy()
    # days with no sales are removed
    df = df[df.y!=0].reset_index(drop=True)
    # detecting outliers with a rolling window std
    anomaly_dict = anomaly_detector(df.y, window_size=10, sigma=2.)
    idx = list(anomaly_dict.keys())
    idx.sort()
    # removing outlier from dataframe
    df.drop(idx, axis=0, inplace=True)
    return df

def train_model(df, kwargs=None, kwargs_extra=None, ts_preproc=True):
    """
    It receives the dataset and Prophet's model parameters 
    and returns a trained instance of the model
    """
    # time-series preprocessing 
    if ts_preproc: df = ts_cleaning(df)
        
    # setting the model parameters
    if kwargs is None: prophet_model = Prophet()
    else: prophet_model = Prophet(**kwargs) 
        
    if kwargs_extra is not None:
        if "wk_prior_scale" in kwargs_extra:
            prophet_model.add_seasonality(name="weekly", period=7, 
                                          fourier_order=kwargs_extra["wk_fourier_order"],
                                          prior_scale=kwargs_extra["wk_prior_scale"])
        if "mt_prior_scale" in kwargs_extra:
            prophet_model.add_seasonality(name='monthly', period=30.5, 
                                          fourier_order=kwargs_extra["mt_fourier_order"],
                                          prior_scale=kwargs_extra["mt_prior_scale"])
        if "yr_prior_scale" in kwargs_extra:
            prophet_model.add_seasonality(name="yearly", period=365.25, 
                                          fourier_order=kwargs_extra["yr_fourier_order"],
                                          prior_scale=kwargs_extra["yr_prior_scale"])
    # model fitting
    prophet_model.fit(df)    
    return prophet_model

def compute_prophet_trend(prophet_model, future_dataframe=None, periods=30, include_history=False):
    if future_dataframe is None:
        future_dataframe = prophet_model.make_future_dataframe(periods=periods, 
                                                               include_history=include_history)
        trend = prophet_model.predict_trend(future_dataframe)
    else:
        future_dataframe = prophet_model.setup_dataframe(future_dataframe.copy())
        trend = prophet_model.predict_trend(future_dataframe)
    return trend

def compute_stl_trend(data, n_periods=0):
    """
    Parameters
    -----------
    data: pandas.DataFrame 
        trainig data with columns 'ds' (dates) and 'y' (values)
    n_periods: int
        number of time periods ahead of data
    Returns
    ----------
    pd.DataFrame
        A dataframe with columns "ds" and "trend" with the trend
        estimation and projection
    """
    # training data in the format of STL
    data = data.copy()
    idx = pd.DatetimeIndex(data.ds) 
    data["date"] = idx
    data.set_index("date", inplace=True)
    data.drop("ds", axis=1, inplace=True)

    # filling gaps
    data_filled = (data
                   .resample("D")
                   .mean()
                   .interpolate("linear"))

    # trend estimation with stl
    stl = decompose(data_filled, period=7)
    stl_trend = stl.trend
    stl_trend.rename(columns={'y':'trend'}, inplace=True)
    stl_fcst = forecast(stl, steps=n_periods, fc_func=drift)
    stl_fcst.rename(columns={'drift':'trend'}, inplace=True)
    stl_trend = stl_trend.append(stl_fcst)

    # just keeping dates with dates in data frame
    idx = idx.append(stl_fcst.index)
    stl_trend = stl_trend.reindex(idx)
    stl_trend.reset_index(inplace=True, drop=True)
    return stl_trend


class TrendEstimator():
    """
    Parameters
    ----------
    backend: string
        name of the backend to be used: ('prophet', ...)
    """

    def __init__(self, backend='prophet'):
        self.backend = backend

    def fit(self, data):
        """
        Parameters
        ----------
        data: pandas.DataFrame
            dataframe with columns 'ds' and 'y'
        """
        trend_estimator = train_model(data,
                            prophet_kwargs, 
                            prophet_kwargs_extra, 
                            ts_preproc=True)
        self.trend_estimator = trend_estimator

    def predict(self, predict_period):
        """
        Parameters
        ----------
        predict_period: pandas.DataFrame
            dataframe with column 'ds' containing the time period to be predicted
        """
        # evaluate over the data_time_range period
        if self.backend == 'prophet':
            trend = compute_prophet_trend(self.trend_estimator, future_dataframe=predict_period)
            trend_dataframe = pd.DataFrame({'ds':predict_period.ds.values, 'trend':trend.values})
        return trend_dataframe




import numpy as np
import pandas as pd
import unittest

from tsforest.forecaster_lightgbm import LightGBMForecaster
from tsforest.forecaster_h2ogbm import H2OGBMForecaster
from tsforest.forecaster_catboost import CatBoostForecaster
from tsforest.grid_search import GridSearch
from tsforest.utils import make_time_range

DATA_PATH = './tests/tests_data/data.csv'

class TestGridSearch(unittest.TestCase):

    def test_it_fit_with_valid_period_in_lightgbm(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        train_data = data.query('ds < "2019-06-01"')
        valid_period = make_time_range('2019-05-01', '2019-05-31', freq='D')
        hyperparams = {'num_leaves':[25,30]}
        hyperparams_fixed = {'num_iterations':30, 'learning_rate':0.3}
  
        gs = GridSearch(model_class=LightGBMForecaster,
                features=['calendar_mixed'], 
                hyperparams=hyperparams,
                hyperparams_fixed=hyperparams_fixed,
                n_jobs=-1)
        gs.fit(train_data=train_data, valid_period=valid_period)

    def test_it_fit_with_test_data_in_lightgbm(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        train_data = data.query('ds < "2019-06-01"')
        test_data = data.query('ds >= "2019-06-01"')
        hyperparams = {'num_leaves':[25,30]}
        hyperparams_fixed = {'num_iterations':30, 'learning_rate':0.3}
        
        gs = GridSearch(model_class=LightGBMForecaster,
                features=['calendar_mixed'], 
                hyperparams=hyperparams,
                hyperparams_fixed=hyperparams_fixed,
                n_jobs=-1)
        gs.fit(data, test_data=test_data)

    def test_it_fit_with_valid_period_in_h2ogbm(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        train_data = data.query('ds < "2019-06-01"')
        valid_period = make_time_range('2019-05-01', '2019-05-31', freq='D')
        hyperparams = {'max_depth':[5,7]}
        hyperparams_fixed = {'ntrees':30, 'learn_rate':0.3}

        gs = GridSearch(model_class=H2OGBMForecaster,
                features=['calendar_mixed'], 
                hyperparams=hyperparams,
                hyperparams_fixed=hyperparams_fixed,
                n_jobs=-1)
        gs.fit(train_data=train_data, valid_period=valid_period)

    def test_it_fit_with_test_data_in_h2ogbm(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        train_data = data.query('ds < "2019-06-01"')
        test_data = data.query('ds >= "2019-06-01"')
        hyperparams = {'max_depth':[5,7]}
        hyperparams_fixed = {'ntrees':30, 'learn_rate':0.3}

        gs = GridSearch(model_class=H2OGBMForecaster,
                features=['calendar_mixed'], 
                hyperparams=hyperparams,
                hyperparams_fixed=hyperparams_fixed,
                n_jobs=-1)
        gs.fit(train_data=train_data, test_data=test_data)

    def test_it_fit_with_valid_period_in_catboost(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        train_data = data.query('ds < "2019-06-01"')
        valid_period = make_time_range('2019-05-01', '2019-05-31', freq='D')
        hyperparams = {'depth':[5,6]}
        hyperparams_fixed = {'iterations':30, 'learning_rate':0.3}
  
        gs = GridSearch(model_class=CatBoostForecaster,
                features=['calendar_mixed'], 
                hyperparams=hyperparams,
                hyperparams_fixed=hyperparams_fixed,
                n_jobs=-1)
        gs.fit(train_data=train_data, valid_period=valid_period)

    def test_it_fit_with_test_data_in_catboost(self):
        data = pd.read_csv(DATA_PATH, parse_dates=['ds'])
        train_data = data.query('ds < "2019-06-01"')
        test_data = data.query('ds >= "2019-06-01"')
        hyperparams = {'depth':[5,6]}
        hyperparams_fixed = {'iterations':30, 'learning_rate':0.3}
        
        gs = GridSearch(model_class=CatBoostForecaster,
                features=['calendar_mixed'], 
                hyperparams=hyperparams,
                hyperparams_fixed=hyperparams_fixed,
                n_jobs=-1)
        gs.fit(data, test_data=test_data)



import numpy as np
import pandas as pd
from zope.interface import implementer
import h2o
from h2o.estimators import H2OGradientBoostingEstimator
# init the cluster if is not already up
if h2o.cluster() is None: h2o.init(nthreads=-1)

from tsforest.config import gbm_parameters
from tsforest.forecaster_base import ForecasterBase
from tsforest.forecaster_interface import ForecasterInterface

@implementer(ForecasterInterface)
class H2OGBMForecaster(ForecasterBase):
    '''
    Parameters
    ----------
    model_params : dict
        Dictionary containing the specific parameters of the boosting model 
    features: list
        List of features to be included
    features_types: dict
        Dictionary containing the type of the features
    detrend: bool
        Whether or not to remove the trend from time serie
    response_scaling:
        Whether or not to perform scaling of the reponse variable
    lags: list
        List of integer lag values
    window_sizes: list
        List of integer window sizes values
    window_functions: list
        List of string names of the window functions
    '''
    def __init__(self, model_params=dict(), features=['calendar_mixed','events'], features_types=dict(),
                 detrend=True, response_scaling=True, lags=None, window_sizes=None, window_functions=None):

        if lags is not None and 'lag' not in features:
            features.append('lag')
        if (window_sizes is not None and window_functions is not None) and 'rw' not in features:
            features.append('rw')

        self.model_params = model_params
        self.features = features
        self.features_types = features_types
        self.detrend = detrend
        self.response_scaling = response_scaling
        self.lags = lags
        self.window_sizes = window_sizes
        self.window_functions = window_functions

    def _cast_dataframe(self, features_dataframe, features_types):
        """
        Parameters
        ----------
        features_dataframe: pandas.DataFrame
            dataframe containing all the features
        features_types: dict
            dictionary containing the type of the features
        Returns
        ----------
        features_dataframe_casted: h2o.H2OFrame
            features dataframe casted to H2O dataframe format
        """
        _features_types = {feat:dtype for feat,dtype in features_types.items()
                           if feat in self.input_features}
        features_dataframe_casted = h2o.H2OFrame(features_dataframe, 
                                                 column_types=_features_types)
        return features_dataframe_casted

    def fit(self, train_data, valid_period=None, early_stopping_rounds=20):
        '''
        Parameters
        ----------
        train_data : pandas.DataFrame
            dataframe with "at least" columns "ds" and "y"
        valid_period: pandas.Series or pandas.DataFrame
            series or dataframe (with column "ds") indicating the validation period
        '''
        assert {"ds","y"} <= set(train_data.columns.values), \
            '"train_data" must contain columns "ds" and "y"'
        
        train_features,features_types = super()._prepare_train_features(train_data)
        if valid_period is not None:
            if isinstance(valid_period, pd.core.series.Series):
                valid_period = pd.DataFrame(valid_period, columns=['ds'])
            valid_features = super()._prepare_valid_features(valid_period, train_features)
            valid_start_time = valid_features.ds.min()
            # removes validation period from train data
            train_data = train_data.query('ds < @valid_start_time')
            train_features = train_features.query('ds < @valid_start_time')

        train_features['y_hat'] = super()._prepare_train_response(train_features)
        if valid_period is not None:
            valid_features['y_hat'] = super()._prepare_valid_response(valid_features)

        train_features_casted = self._cast_dataframe(train_features, features_types)
        valid_features_casted = self._cast_dataframe(valid_features, features_types) \
                                if valid_period is not None else None

        self.train_data = train_data
        self.features_types = features_types
        self.train_features = train_features
        self.train_features_casted = train_features_casted

        self.valid_period = valid_period
        self.valid_features = valid_features if valid_period is not None else None
        self.valid_features_casted = valid_features_casted

        # model_params overwrites default params of model
        model_params = {**gbm_parameters, **self.model_params}

        training_params = {'training_frame':train_features_casted, 
                           'x':self.input_features, 
                           'y':self.target}
        if valid_period is not None:
            training_params['validation_frame'] = valid_features_casted
            model_params['stopping_rounds'] = early_stopping_rounds 
        if 'weight' in self.train_features.columns:
            training_params['weights_column']='weight'

        # model training
        model = H2OGradientBoostingEstimator(**model_params)
        model.train(**training_params)
        self.model = model
        self.best_iteration = int(model.summary()["number_of_trees"][0])

    def _predict(self, model, test_features, trend_dataframe):
        """
        Parameters
        ----------
        model: LightGBM 
            Trained model
        test_features: pandas.DataFrame
            datafame containing the features for the test period
        trend_dataframe: pandas.DataFrame
            dataframe containing the trend estimation over the test period
        """
        y_train = self.train_features.y.values
        y_valid = self.valid_features.y.values \
                  if self.valid_period is not None else np.array([])
        y = np.concatenate([y_train, y_valid])

        prediction = list()
        for idx in range(test_features.shape[0]):
            if 'lag' in self.features:
                for lag in self.lags:
                    test_features.loc[idx, f'lag_{lag}'] = y[-lag]
            if 'rw' in self.features:
                for window_func in self.window_functions:
                    for window in self.window_sizes:
                        test_features.loc[idx, f'{window_func}_{window}'] = getattr(np, window_func)(y[-window:])
            test_features_casted = self._cast_dataframe(test_features.loc[[idx], :], 
                                                        self.features_types)
            _y_pred = model.predict(test_features_casted)
            y_pred = _y_pred.as_data_frame().values[:,0]
            prediction.append(y_pred.copy())
            if self.response_scaling:
                y_pred *= self.y_std
                y_pred += self.y_mean
            if self.detrend:
                y_pred += trend_dataframe.loc[idx, 'trend']
            y = np.append(y, [y_pred])
        return np.asarray(prediction).ravel()

    def predict(self, test_period):
        '''
        Parameters
        ----------
        test_period: pandas.DataFrame
            dataframe with the same columns as "train_data" except for "y"
        Returns
        ----------
        prediction_dataframe: pandas.DataFrame
            dataframe containing dates "ds" and predictions "y_pred"
        '''
        assert set(self.train_data.columns) - set(test_period.columns) == {'y'}, \
            '"test_period" must have the same columns as self.train_data except for "y"'
        
        test_features = super()._prepare_test_features(test_period)
        if self.detrend:
            trend_estimator = self.trend_estimator
            trend_dataframe = trend_estimator.predict(test_period.loc[:, ['ds']])
        else:
            trend_dataframe = None

        if 'lag' in self.features or 'rw' in self.features:
            prediction = self._predict(self.model, test_features, trend_dataframe)
        else:
            test_features_casted = self._cast_dataframe(test_features, 
                                                        self.features_types)
            _prediction = self.model.predict(test_features_casted)
            prediction = _prediction.as_data_frame().values[:,0]

        if self.response_scaling:
            prediction *= self.y_std
            prediction += self.y_mean
        if self.detrend:
            prediction += trend_dataframe.trend.values
        if "closed" in test_features.columns:
            closed_mask = test_features["closed"]==1
            prediction[closed_mask] = 0
        
        self.test_features = test_features
            
        prediction_dataframe = pd.DataFrame({"ds":test_period.ds, "y_pred":prediction})
        return prediction_dataframe

    def show_variable_importance(self):
        pass

    def save_model(self):
        pass


import pandas as pd
from inspect import getmembers, isfunction

from tsforest.config import gbm_parameters
from tsforest.features import FeaturesGenerator
from tsforest.trend import TrendEstimator
from tsforest import metrics

class ForecasterBase(object):
      
    def _prepare_train_features(self, train_data):
        '''
        Parameters
        ----------
        train_data : pandas.DataFrame
            dataframe with at least columns "ds" and "y"
        '''
        features_generator = FeaturesGenerator(include_features=self.features,
                                               lags=self.lags,
                                               window_sizes=self.window_sizes,
                                               window_functions=self.window_functions)
        train_features,features_types = features_generator.compute_train_features(train_data)
        features_types = {**features_types, **self.features_types}

        exclude_features = ['ds', 'y', 'y_hat', 'month_day', 'weight', 'fold_column']
        self.input_features = [feature for feature in train_features.columns
                               if feature not in exclude_features]
        self.features_generator = features_generator
        return train_features,features_types
    
    def _prepare_valid_features(self, valid_period, train_features):
        '''
        valid_period : pandas.DataFrame
            dataframe with column "ds" indicating the validation period
        train_features: pandas.DataFrame
            dataframe
        '''
        valid_features = pd.merge(valid_period, train_features, how='inner', on=['ds'])
        assert len(valid_features)==len(valid_period), \
            'valid_period must be contained in the time period of time_features'
        return valid_features

    def _prepare_test_features(self, test_period):
        '''
        Parameters
        ----------
        test_data: pandas.DataFrame
            dataframe with the same columns as self.train_data (except for "y") 
            containing the test period
        Returns
        ----------
        test_features: pandas.DataFrame
            Dataframe containing all the features for evaluating the trained model
        '''
        test_features,features_types = self.features_generator.compute_test_features(test_period)
        return test_features
    
    def _prepare_train_response(self, train_features):
        '''
        Prepares the train response variable

        Parameters
        ----------
        train_features: pd.DataFrame
            dataframe containing the columns "ds" and "y"
        '''
        y_hat = train_features.y.copy()

        if self.detrend:
            trend_estimator = TrendEstimator()
            trend_estimator.fit(data=train_features.loc[:, ['ds','y']])
            trend_dataframe = trend_estimator.predict(train_features.loc[:, ['ds']])
            y_hat -= trend_dataframe.trend.values
            self.trend_estimator = trend_estimator
        if self.response_scaling:
            y_mean = y_hat.mean()
            y_std = y_hat.std()
            y_hat -= y_mean
            y_hat /= y_std 

        self.y_mean = y_mean if 'y_mean' in locals() else None
        self.y_std  = y_std if 'y_std' in locals() else None
        self.target = 'y_hat'
        return y_hat
    
    def _prepare_valid_response(self, valid_features):
        '''
        Prepares the validation response variable

        Parameters
        ----------
        valid_features: pd.DataFrame
            dataframe containing the columns "ds" and "y"
        '''
        y_hat = valid_features.y.copy()

        if self.detrend:
            trend_estimator = self.trend_estimator
            trend_dataframe = trend_estimator.predict(valid_features.loc[:, ['ds']])
            y_hat -= trend_dataframe.trend.values
        if self.response_scaling:
            y_hat -= self.y_mean
            y_hat /= self.y_std 
            
        return y_hat

    def evaluate(self, test_data, metric='rmse'):
        '''
        Parameters
        ----------
        test_data: pandas.DataFrame
            dataframe with the same columns as "train_data"
        metric: string
            possible values: "mae", "mape", "mse", "rmse", "smape"
        Returns
        ----------
        error: float
            error of predictions according to the error measure
        '''
        assert set(self.train_data.columns) == set(test_data.columns), \
            '"test_data" must have the same columns as "train_data"'
        available_metrics = [member[0].split('_')[1] for member in getmembers(metrics) 
                             if isfunction(member[1])]
        assert metric in available_metrics, \
            f'"metric" must be any of these: {available_metrics}'
        test_data = test_data.copy()
        y_real = test_data.pop("y")
        y_pred = self.predict(test_data)["y_pred"].values
        error_func = getattr(metrics, f'compute_{metric}')
        error = error_func(y_real, y_pred)
        return error
